/*
 * nqsd/nqs_enf.c
 * Enforce kernel-based resource limits
 * 
 * DESCRIPTION:
 *
 *	This module enforces batch request resource limits.
 *
 *	Original Author:
 *	-------
 *	Robert W. Sandstrom, Sterling Software Incorporated.
 *	February 17, 1986.
 */

#include <libnqs/license.h>
#include <libnqs/proto.h>

#include <errno.h>
#include <libnqs/informcc.h>		/* NQS information completion */
					/* codes and masks */
#include <libnqs/requestcc.h>		/* NQS request completion codes */
#include <unistd.h>

#if	SYS52 
#else
#if 	UNICOS
#include <time.h>
#include <sys/category.h>
#include <sys/resource.h>
#else
#if	IBMRS
#include <sys/sched.h>
#include <sys/time.h>			/* So we can include sys/resource.h */
#include <sys/resource.h>		/* Get definition of struct rlimit */
#else
#if	HPUX
#define RLIM_INFINITY  0x7fffffff
#define RLIMIT_CPU	0
#define RLIMIT_FSIZE	1
#define RLIMIT_DATA     2               /* data size */
#define RLIMIT_STACK    3               /* stack size */
#define RLIMIT_CORE     4               /* core file size */
#define RLIMIT_RSS      5               /* resident set size */
#include <sys/resource.h>
#include <sys/time.h>
#else
#if	POSIX | SYSVr4 | BSD
#include <sys/time.h>			/* So we can include sys/resource.h */
#include <sys/resource.h>		/* Get definition of struct rlimit */
#else
BAD SYSTEM TYPE
#endif
#endif
#endif
#endif
#endif

#if	VALID_LIMITS & (LIM_PPCPUT | LIM_PRCPUT)
#if	POSIX | SYSVr4 | BSD
#if	CPU_LIMIT_GRAN > 2000
REWRITE THIS FUNCTION
/* See the note about cpulimhigh() below. */
#endif
/*** enf_bsdcpu
 *
 *
 *	void enf_bsdcpu():
 *
 *	Enforce within BSD4.2 a per process or per request
 *	cpu time limit.
 */
void enf_bsdcpu (
	struct cpulimit *cpulimitp,	/* Pointer to what is to be set */
	short infinite,			/* Boolean */
	long limit_type)		/* Per_process or Per_request */
{
	int limresult;
	struct rlimit rlimit;
	
	if (limit_type == LIM_PPCPUT) {
		if (infinite) {
			/*
			 * It is impossible in NQS to have an infinite
			 * maximum and a finite warning.
			 */
			rlimit.rlim_max = RLIM_INFINITY;
			rlimit.rlim_cur = RLIM_INFINITY;
		}
		else {
			/*
			 * If seconds > 1000, ignore ms in order to avoid
			 * overflow.  (1000 is arbitrary. We know from
			 * cpulimhigh() that the limit can be enforced;
			 * we just have to avoid ever expressing the limit
			 * in units finer than those of the system call.)
			 */
			if (cpulimitp->max_seconds > 1000) {
				rlimit.rlim_max =
					cpulimitp->max_seconds * CPU_LIMIT_GRAN;
			}
			/*
			 * It is safe to temporarily express the limit
			 * in units 1000 times finer than the system call.
			 */
			else {
				rlimit.rlim_max =
					(((1000 * cpulimitp->max_seconds)
					+ cpulimitp->max_ms) *
					CPU_LIMIT_GRAN) / 1000;
			}
			/* Same reasoning as above. */
			if (cpulimitp->warn_seconds > 1000) {
				rlimit.rlim_cur =
					cpulimitp->warn_seconds *
					CPU_LIMIT_GRAN;
			}
			else {
				rlimit.rlim_cur =
					(((1000 * cpulimitp->warn_seconds)
					+ cpulimitp->warn_ms) *
					CPU_LIMIT_GRAN) / 1000;
			}
		}
		limresult = setrlimit (RLIMIT_CPU, &rlimit);
		if (limresult == -1L) {
			serexit (RCM_UNABLETOEXE | TCI_PP_CTLEXC,
				 asciierrno());
		}
	}
}
#else
#if	UNICOS

void enf_bsdcpu (
	struct cpulimit *cpulimitp,	/* Pointer to what is to be set */
	short infinite,			/* Boolean */
	long limit_type)		/* Per_process or Per_request */
{
  	long newlimit;
  
  	if (infinite)
    		newlimit = CPUUNLIM;
  	else {
    		newlimit = cpulimitp->max_seconds * CLK_TCK;
	}
  	
  	if (limit_type == LIM_PPCPUT) {
	  	limit(C_PROC, 0, L_CPU, newlimit);
	}
  	else if (limit_type == LIM_PRCPUT) {
	  	limit(C_JOB, 0, L_CPU, newlimit);
	}
}
#else
#if	SYS52 
#else
BAD SYSTEM TYPE
#endif
#endif
#endif
#endif


#if	VALID_LIMITS & LIM_PPNICE
#if	BSD42 | BSD43 | ULTRIX | DECOSF | IBMRS
/*** enf_bsdnic
 *
 *
 *	void enf_bsdnic():
 *
 *	Enforce a nice value within BSD4.2.
 */
void enf_bsdnic (int nice_value)
{
	int limresult;
	
	limresult = setpriority (PRIO_PROCESS, getpid(), nice_value);
	if (limresult == -1) {
		serexit (RCM_UNABLETOEXE | TCI_PP_NELEXC, asciierrno());
	}
}
#else
#if	HPUX | SGI | SOLARIS | SYS52 | LINUX | UNICOS
/*** enf_sy5nic
 *
 *
 *	void enf_sy5nic():
 *
 *	Enforce a nice value within Silicon Graphics' Unix, System V,
 *
 */
void enf_sy5nic (int nice_value)
{
	int limresult;
	
	errno = 0;			/* nice(2) returning -1 doesn't */
					/* always mean failure */
	limresult = nice (0);		/* Get current nice value */
	if (errno != 0) {
		serexit (RCM_UNABLETOEXE | TCI_PP_NELEXC, asciierrno());
	}
	nice (nice_value - limresult);    /* Go to desired value */
	if (errno != 0) {
		serexit (RCM_UNABLETOEXE | TCI_PP_NELEXC, asciierrno());
	}
#if	0
#if	IBMRS
	/*
	 * As a result of this, nice 16 becomes priority 121 and nice 19
	 * becomes priority 124.
	 */
	if (nice_value > 15) {
	    nice_value += PRIORITY_MAX - 21;
	    setpri (0, nice_value);
	}
#endif
#endif
}
#else
BAD SYSTEM TYPE
#endif
#endif
#endif


/*
 * Below, we define THESE, THEM, and THOSE for C preprocessors that
 * do not allow line continuation.
 */
#define THESE	LIM_PPCORE | LIM_PPDATA | LIM_PPPFILE | LIM_PRPFILE
#define THEM	LIM_PPMEM | LIM_PRMEM | LIM_PPQFILE | LIM_PRQFILE 
#define THOSE	LIM_PPSTACK | LIM_PPTFILE | LIM_PRTFILE | LIM_PPWORK

#if	VALID_LIMITS & (THESE | THEM | THOSE)
/*** enf_bsdquo
 *
 *
 *	void enf_bsdquo():
 *
 *	Enforce within BSD4.2 one of the limits: per process core file
 *	size, per process data segment size, per process permanent file
 *	size, per request permanent file size, per process
 *	memory size, per request memory size, per process stack
 *	segment size, per process temporary file size, per request
 *	temporary file size, or per process working set size.
 *
 */
void enf_bsdquo (
	struct quotalimit *quotalimitp,	/* Pointer to what is to be set */
	short infinite,			/* Boolean */
	long limit_type)		/* One of LIM_??? */
{

#if 	UNICOS
	long newlimit;
	int granularity;

	if (infinite)
		newlimit = 0;
	else {
		granularity = 4096;  /* click */
		switch (quotalimitp->max_units) {
		case QLM_BYTES:
			newlimit = quotalimitp->max_quota / granularity;
			break;
		case QLM_WORDS:
			newlimit =
				(quotalimitp->max_quota * BYTES_PER_WORD) /
				granularity;
			break;
		case QLM_KBYTES:
			newlimit = (quotalimitp->max_quota * (1<<10)) /
				granularity;
			break;
		case QLM_KWORDS:
			newlimit =
				(quotalimitp->max_quota * (BYTES_PER_WORD) *
				(1<<10)) / granularity;
			break;
		case QLM_MBYTES:
			newlimit = (quotalimitp->max_quota * (1<<20)) /
				granularity;
			break;
		case QLM_MWORDS:
			newlimit =
				(quotalimitp->max_quota * (BYTES_PER_WORD) *
				(1<<20)) / granularity;
			break;
		case QLM_GBYTES:
			newlimit = (quotalimitp->max_quota * (1<<30)) /
				granularity;
			break;
		case QLM_GWORDS:
			newlimit =
				(quotalimitp->max_quota * (BYTES_PER_WORD) *
				(1<<30)) / granularity;
			break;
		}
	}

	switch (limit_type) {
	case LIM_PPCORE:
		limit(C_PROC, 0, L_CORE, newlimit);
		break;
	case LIM_PPMEM:
		limit(C_PROC, 0, L_MEM, newlimit);
		break;
	case LIM_PRMEM:
		limit(C_JOB, 0, L_MEM, newlimit);
		break;
	case LIM_PPPFILE:
		limit(C_PROC, 0, L_FSBLK, newlimit);
		break;
	case LIM_PRPFILE:
		limit(C_JOB, 0, L_FSBLK, newlimit);
		break;
	}

#else
	int granularity;		/* Granularity of the sys call */
	int limresult;			/* Return value of limit setting */
	struct rlimit rlimit;		/* Holds the limit itself */

	if (infinite) {
		/*
		 * It is impossible in NQS to have an infinite maximum
		 * and a finite warning.
		 */
		rlimit.rlim_max = RLIM_INFINITY;
		rlimit.rlim_cur = RLIM_INFINITY;
	}
	else {
		if (limit_type & (LIM_PPCORE | LIM_PPPFILE)) {
			granularity = DISK_LIMIT_GRAN;
		}
		else {
			granularity = MEM_LIMIT_GRAN;
		}
		switch (quotalimitp->max_units) {
		case QLM_BYTES:
			rlimit.rlim_max = quotalimitp->max_quota / granularity;
			break;
		case QLM_WORDS:
			rlimit.rlim_max =
				(quotalimitp->max_quota * BYTES_PER_WORD) /
				granularity;
			break;
		case QLM_KBYTES:
			rlimit.rlim_max = (quotalimitp->max_quota * (1<<10)) /
				granularity;
			break;
		case QLM_KWORDS:
			rlimit.rlim_max =
				(quotalimitp->max_quota * (BYTES_PER_WORD) *
				(1<<10)) / granularity;
			break;
		case QLM_MBYTES:
			rlimit.rlim_max = (quotalimitp->max_quota * (1<<20)) /
				granularity;
			break;
		case QLM_MWORDS:
			rlimit.rlim_max =
				(quotalimitp->max_quota * (BYTES_PER_WORD) *
				(1<<20)) / granularity;
			break;
		case QLM_GBYTES:
			rlimit.rlim_max = (quotalimitp->max_quota * (1<<30)) /
				granularity;
			break;
		case QLM_GWORDS:
			rlimit.rlim_max =
				(quotalimitp->max_quota * (BYTES_PER_WORD) *
				(1<<30)) / granularity;
			break;
		}
		switch (quotalimitp->warn_units) {
		case QLM_BYTES:
			rlimit.rlim_cur = quotalimitp->warn_quota / granularity;
			break;
		case QLM_WORDS:
			rlimit.rlim_cur =
				(quotalimitp->warn_quota * BYTES_PER_WORD) /
				granularity;
			break;
		case QLM_KBYTES:
			rlimit.rlim_cur = (quotalimitp->warn_quota * (1<<10)) /
				granularity;
			break;
		case QLM_KWORDS:
			rlimit.rlim_cur =
				(quotalimitp->warn_quota * (BYTES_PER_WORD) *
				(1<<10)) / granularity;
			break;
		case QLM_MBYTES:
			rlimit.rlim_cur = (quotalimitp->warn_quota * (1<<20)) /
				granularity;
			break;
		case QLM_MWORDS:
			rlimit.rlim_cur =
				(quotalimitp->warn_quota * (BYTES_PER_WORD) *
				(1<<20)) / granularity;
			break;
		case QLM_GBYTES:
			rlimit.rlim_cur = (quotalimitp->warn_quota * (1<<30)) /
				granularity;
			break;
		case QLM_GWORDS:
			rlimit.rlim_cur =
				(quotalimitp->warn_quota * (BYTES_PER_WORD) *
				(1<<30)) / granularity;
			break;
		}
	}
	switch (limit_type) {
	case LIM_PPCORE:
		limresult = setrlimit (RLIMIT_CORE, &rlimit);
		if (limresult == -1) {
			serexit (RCM_UNABLETOEXE | TCI_PP_CFLEXC,
				 asciierrno());
		}
		break;
	case LIM_PPDATA:
		limresult = setrlimit (RLIMIT_DATA, &rlimit);
		if (limresult == -1) {
			serexit (RCM_UNABLETOEXE | TCI_PP_DSLEXC,
				 asciierrno());
		}
		break;
	case LIM_PPPFILE:
		limresult = setrlimit (RLIMIT_FSIZE, &rlimit);
		if (limresult == -1) {
			serexit (RCM_UNABLETOEXE | TCI_PP_PFLEXC,
				 asciierrno());
		}
		break;
	case LIM_PPSTACK:
		limresult = setrlimit (RLIMIT_STACK, &rlimit);
		if (limresult == -1) {
			serexit (RCM_UNABLETOEXE | TCI_PP_SSLEXC,
				 asciierrno());
		}
		break;
#if	(VALID_LIMITS & LIM_PPWORK)
	case LIM_PPWORK:
		limresult = setrlimit (RLIMIT_RSS, &rlimit);
		if (limresult == -1) {
			serexit (RCM_UNABLETOEXE | TCI_PP_WSLEXC,
				 asciierrno());
		}
		break;
#endif
#if	(VALID_LIMITS & LIM_PPMEM)
	case LIM_PPMEM:
		limresult = setrlimit (RLIMIT_VMEM, &rlimit);
		if (limresult == -1) {
			serexit (RCM_UNABLETOEXE | TCI_PP_MSLEXC,
				 asciierrno());
		}
		break;
#endif
	}
#endif
}
#endif
